#import structures

# code for parsing commands, dags, trees and automata

from machines import BDT, Rule
from structures import Tree, Dag
import sys
from pyparsing import Word, alphas, nums, alphanums, Suppress, Literal, delimitedList, Forward, Optional, Empty, FollowedBy, ZeroOrMore, Group, OneOrMore, printables, oneOf

TREE = ("tree",)
PENMAN = ("penman", "p")
DAG = ("dag", "graph", "g")
DERIVATIONS = ("derivations", "d")
AUTOMATON = ("automaton", "acceptor", "a")
WEIGHTED = ("weighted", "w")
EXTENDED = ("extended", "x")
TRANSDUCER = ("transducer", "t")
EVALUATE = ("evaluate", "e", "r", "apply")
NBEST = ("nbest", "n", "kbest", "k")
FILE = ("file", "f")
STDIN = ("stdin", "s")
QUERY = ("info", "query", "q")
INTERSECT = ("intersection", "intersect", "i")
UNION = ("union", "u")
CASCADE = ("cascade", "c")
VISUALIZE = ("visualize", "v")
BACKWARD = ("backward", "b")


def dag_from_tuple(t):
    (root, edges, nodes) = t
    d = Dag()
    for (nid, nlabel) in nodes.iteritems():
        d.add_vertex(vid = nid, vlabel = nlabel) #.split(":")[0])
    for (head, pos, tail) in edges:
        role = pos #role = nodes[head].split(":")[pos + 1]
        d.add_edge(head, tail, role)
    d.root = root
    return d
        
def dag_from_penman(s):
#    print s
    t = penman.parseString(s)
    return dag_from_tuple(t[0])
#    print (root, edges, nodes)

def dag_from_file(f):
    return dag_from_penman(f.readline().strip())

penman = Forward()

symbol = Word(alphanums + "_=")
dot = Suppress(Literal("."))
colon = Suppress(Literal(":"))
lbr = Suppress(Literal("("))
rbr = Suppress(Literal(")"))
slash = Suppress(Literal("/"))
ot = Suppress(Literal("<-"))
to = Suppress(Literal("->"))
at = Suppress(Literal("@"))



def make_expression(tok):
    try:
        t, d = (tok[0], tok[1:])
        if t in DAG:
            if d[0][0] in PENMAN:
                return d[0][1][0]
            if d[0][0] in FILE:
                return dag_from_file(open(d[0][1][0]))
        if t in AUTOMATON:
            w = True if (WEIGHTED in d[0]) else False
            t = True if (TRANSDUCER in d[0]) else False
            if d[1][0] in FILE:
                a = bda_from_file(open(d[1][1][0]), weighted=w, transducer=t)
                return a
        if t in EVALUATE:
            dag = d[1]
            dag.add_pseudoroot()
#            return d[0].evaluate(dag)
            return list((s, w, t, t.frontier()) for ((s, t), w) in d[0].evaluate(dag)['results'])
        if t in DERIVATIONS:
            dag = d[1]
            dag.add_pseudoroot()
            return d[0].evaluate(dag)['derivations']
        if t in NBEST:
            return d[1].generate(number=int(d[0]))
        if t in INTERSECT:
            return d[0].intersection(d[1])
        if t in UNION:
            return d[0].union(d[1])
        if t in VISUALIZE:
            dot = d[0].dot()
            dot.write(d[1])
            dot.write_pdf("%s.pdf" % (d[1],))
            return "Written dag/automaton to %s." % (d[1],)
        if t in TREE:
            return d[0]
        if t in QUERY:
            return str(d[0])
        if t in BACKWARD:
            return d[1].backward(d[0], d[2])
        return t, d
    except Exception as e:
#        print e
        return e


def Expression(*a):
    p = lbr + oneOf(a[0])
    for e in a[1:]:
        p += e
    p += rbr
    return p.setParseAction(make_expression)

def Features(l):
    f = {"type": "features"}
    for d in l:
        f.update(d)
    return f

path = Word(alphanums + "-_/.")
automaton = Forward()

Feature = lambda a: oneOf(a).setParseAction(lambda _: dict([(a, True)]))

weighted = Feature(WEIGHTED)
extended = Feature(EXTENDED)
transducer = Feature(TRANSDUCER)
features = ZeroOrMore(weighted ^ extended ^ transducer).setParseAction(Features)

automaton = Forward()

pman = Expression(PENMAN, penman.copy().setParseAction(
    lambda t: dag_from_tuple(t[0])))
stdin = Expression(STDIN)
infile = Expression(FILE, path)
inputs = (stdin ^ infile ^ pman)
tree = Forward()
dag = Expression(DAG, inputs)
simple_automaton = Expression(AUTOMATON, (features + inputs))
objects = (dag ^ automaton ^ tree)
info = Expression(QUERY, objects)
treeexpression = Expression(TREE, tree)


intersect = Expression(INTERSECT, (automaton + automaton))
union = Expression(UNION, (automaton + automaton))
cascade = Expression(CASCADE, OneOrMore(automaton))
evaluate = Expression(EVALUATE, (automaton + dag))
derivations = Expression(DERIVATIONS, (automaton + dag))
visualize = Expression(VISUALIZE, (dag ^ automaton) + path)

automaton << (intersect ^ union ^ cascade ^ simple_automaton)

kbest = Expression(NBEST, Word(nums) + automaton)
backward = Expression(BACKWARD, (Word(nums) + automaton + treeexpression))
result = ZeroOrMore(kbest ^ backward ^ visualize ^ info ^ evaluate ^ derivations ^ objects)
    
def make_tree(tok):
    return Tree(tok[0], tok[1:])

weight = Word(nums + "+-.e")
state = symbol
variable = Word(alphanums)
role = colon + Word(alphanums)
only_variable = Group(variable + colon)
state_variable = Group(state + dot + variable)
treenode = (symbol ^ only_variable ^ state_variable)
trees = ZeroOrMore(tree ^ treenode)
subtrees = (lbr + trees + rbr)
tree << (treenode + Optional(subtrees)).setParseAction(make_tree)


tiburon_line = variable ^ (tree + to + tree)
state_tree = Group(state + dot + tree)
state_opttree = state_tree | Group(state)
state_optvar = state_variable | Group(state)

dag_pattern = lbr + symbol + Group(ZeroOrMore(
    Group(role + state_optvar))) + rbr

bda_lhs = Group(OneOrMore(state_opttree))
bda_rhs = (dag_pattern ^ state) + Optional(at + weight)
bda_line = bda_lhs + to + bda_rhs



def tree_from_string(s):
    return tree.parseString(s)[0]


def rule_from_string(s):
    tokens = bda_line.parseString(s)
    if len(tokens) == 3:
        (lhs, pattern, rhs) = tokens
        weight = 1.0
    elif len(tokens) == 4:
        (lhs, pattern, rhs, weight) = tokens
    else:
        raise ValueError("BDA rule has wrong format!")
    return Rule(lhs, pattern, rhs.asList(), weight=weight)

   

def bda_from_file(f, transducer=False, weighted=False):
    i = f.readline().strip()
    r = []
    for line in f:
        line = line.strip()
        if line and not line.startswith('#'):
            rule = rule_from_string(line)
            r.append(rule)
    bda = BDT(rule_list=r, transducer=transducer, weighted=weighted)
    bda.initial_state = i
    return bda
    
def make_dag_identifier(tok):
    identifier = tok[0]
    if identifier == "-":
        return (0, [], {})
    else:
        return (identifier, [], {})

def make_dag_node(tok):
    (identifier, symbol) = tok
    return (identifier[0], symbol)
   
def make_dag(tok):
    (node, children) = tok
    if children:
        (roles, dags) = zip(*children)
    else:
        (roles, dags) = ([], [])
    (identifier, symbol) = node
    new_symbol = "%s:%s" % (symbol, ":".join(roles))
    new_nodes = {identifier: new_symbol}
    new_edges = []
    for (i, (head, edges, nodes))  in enumerate(dags): 
        new_nodes.update(nodes)
        new_edges.append((identifier, roles[i], head))
        new_edges.extend(edges)
    return (identifier, new_edges, new_nodes)

identifier = Word(nums).setParseAction(make_dag_identifier)
no_node = Literal("-").setParseAction(make_dag_identifier)
node = (Word(nums) + slash + symbol).setParseAction(make_dag_node) 
child = Group(role + penman)
children = Group(ZeroOrMore(child))
mother_children = lbr + (node + children).setParseAction(make_dag) + rbr
penman << (mother_children ^ identifier ^ no_node)

if __name__ == "__main__":
    print result.parseString(sys.argv[1])[0]
